#lang racket
(require "mascheme.rkt")

;; simple representation of a module
;; = function + membrane

(define-struct openModule (fun membrane))

;; cache data structure 

(define cacheStruct (make-hash))

(define (inCache key hash)
  (hash-has-key? hash key))

(define (lookupCache key hash)
  (hash-ref hash key))

(define (updateCache key value hash)
	(hash-set! hash key value))


;; the Math module

(define Math
  
  (let* ((fib (lambda (n) 1))
         (int_memb (new-membrane))
         (ext_memb (new-membrane (call fib))))
  
    ;; Fibonacci recusion aspect

    (define (((fib_rec proceed) . ctx) . args)
      (let ((arg (first args)))
        (if (> arg 2) 
            (+ (fib (- arg 1)) (fib (- arg 2)))
            (apply proceed args)
            )))
    
    ;; caching aspect 

    (define (((cache proceed) . ctx) . args)
      (let ((arg (first args)))        
        (if (inCache arg cacheStruct)
            (lookupCache arg cacheStruct)
            (let ((value (apply proceed args)))  
              (begin (updateCache arg value cacheStruct))
              value)
            )))
    
    (fib_rec (lambda arg (fib_rec arg)))
    (advise int_memb int_memb) 
    (register (call fib) fib_rec int_memb) 
    (register (call fib) cache int_memb) 
    (openModule
     (lambda (n)
       (deploy-fluid int_memb
                     (deploy-fluid ext_memb 
                                   (fib n))))
     ext_memb)))

;; logging aspect

(define (((logging proceed) . ctx) . args)
 (let ((value (apply proceed args)))    
   (printf "~a -> ~a is logged~n" args value)))


;; importing the module Math

(let ((fib      (openModule-fun Math))   
      (fib_memb (openModule-membrane Math)))
  ;; advising it
  (let ((log_memb (new-membrane)))
    (register call? logging log_memb)
    (advise log_memb fib_memb)
    ;; using fib
    (fib 10)))
